"""Processing of systems."""

from argparse import Namespace
from datetime import datetime
from multiprocessing.managers import DictProxy
from typing import NamedTuple, Tuple

from homeinfotools.logging import syslogger
from homeinfotools.rpc.exceptions import SSHConnectionError
from homeinfotools.rpc.reboot import reboot
from homeinfotools.rpc.runcmd import runcmd
from homeinfotools.rpc.sysupgrade import sysupgrade


__all__ = ['Worker']


class Worker(NamedTuple):
    """Stored args and manager to process systems."""

    args: Namespace

    def __call__(self, system: int) -> Tuple[int, DictProxy]:
        """Runs the worker on the given system."""
        start = datetime.now()
        result = {'start': start.isoformat()}
        success = True

        try:
            if self.args.sysupgrade:
                result['sysupgrade'] = sysupgrade(system, self.args)

            if self.args.execute:
                result['execute'] = runcmd(system, self.args)

            if self.args.reboot:
                result['reboot'] = reboot(system, self.args)
        except SSHConnectionError:
            syslogger(system).error('Could not establish SSH connection.')

        result['success'] = success
        end = datetime.now()
        result['end'] = end.isoformat()
        result['duration'] = str(end - start)
        return (system, result)



"""Common constants."""

from homeinfotools.os import SSH


__all__ = [
    'HOSTNAME',
    'PACMAN',
    'SUDO',
    'SSH',
    'SSH_OPTIONS',
    'SYSTEMCTL'
]


HOSTNAME = '{}.terminals.homeinfo.intra'
PACMAN = '/usr/bin/pacman'
SSH_OPTIONS = (
    'LogLevel=error', 'UserKnownHostsFile=/dev/null',
    'StrictHostKeyChecking=no', 'ConnectTimeout=5'
)
SUDO = '/usr/bin/sudo'
SYSTEMCTL = '/usr/bin/systemctl'



"""Terminal backtch updating utility."""

from json import dump
from logging import basicConfig
from multiprocessing import Pool

from homeinfotools.logging import LOG_FORMAT
from homeinfotools.rpc.argparse import get_args
from homeinfotools.rpc.functions import get_log_level
from homeinfotools.rpc.processing import Worker


__all__ = ['main']


def main():
    """Runs the script."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=get_log_level(args))

    with Pool(processes=args.processes) as pool:
        result = pool.map(Worker(args), args.system)

    if args.json is not None:
        with args.json.open('w') as file:
            dump(dict(result), file, indent=2)



"""Argument parsing."""

from argparse import ArgumentParser, Namespace
from pathlib import Path


__all__ = ['get_args']


def get_args() -> Namespace:
    """Returns parsed CLI arguments."""

    parser = ArgumentParser(description='Batch upgrade systems.')
    parser.add_argument('system', type=int, nargs='+',
                        help='systems to upgrade')
    parser.add_argument('-S', '--sysupgrade', action='store_true',
                        help='upgrade the systems')
    parser.add_argument('-X', '--execute', metavar='command',
                        help='execute the commands on the systems')
    parser.add_argument('-R', '--reboot', action='store_true',
                        help='reboot the systems')
    parser.add_argument('-c', '--cleanup', action='store_true',
                        help='cleanup unneeded packages after upgrade')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='enable debug logging')
    parser.add_argument('-i', '--install', default=(), nargs='+',
                        metavar='pkg', help='install the given packages')
    parser.add_argument('-j', '--json', type=Path, metavar='file',
                        help='log jobs as JSON when done')
    parser.add_argument('-k', '--keyring', action='store_true',
                        help='upgrade the keyring before sysupgrade')
    parser.add_argument('-n', '--no-stdin', action='store_true',
                        help='make ssh to not read STDIN')
    parser.add_argument('-o', '--overwrite', default=(), nargs='+',
                        metavar='glob', help='globs of files to overwrite')
    parser.add_argument('-p', '--processes', type=int, metavar='n',
                        help='amount of parallel processes')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='enable verbose logging')
    parser.add_argument('-y', '--yes', action='store_true',
                        help='assume yes on program queries')
    return parser.parse_args()



"""System upgrade."""

from argparse import Namespace
from subprocess import CompletedProcess

from homeinfotools.logging import syslogger
from homeinfotools.rpc.common import PACMAN
from homeinfotools.rpc.exceptions import SSHConnectionError
from homeinfotools.rpc.exceptions import PacmanError
from homeinfotools.rpc.exceptions import SystemIOError
from homeinfotools.rpc.exceptions import UnknownError
from homeinfotools.rpc.functions import completed_process_to_json
from homeinfotools.rpc.functions import execute
from homeinfotools.rpc.functions import ssh
from homeinfotools.rpc.functions import sudo


__all__ = ['sysupgrade']


def lograise(system: int, message: str, completed_process: CompletedProcess):
    """Issues a warning message and raises an exception."""

    if completed_process.returncode == 255:
        raise SSHConnectionError(completed_process)

    # Do not warn on SSH connection errors.
    syslogger(system).warning(message)

    if completed_process.returncode == 126:
        raise SystemIOError(completed_process)

    if completed_process.returncode == 1:
        raise PacmanError(completed_process)

    raise UnknownError(completed_process)


def upgrade_keyring(system: int, args: Namespace) -> CompletedProcess:
    """Upgrades the archlinux-keyring on that system."""

    command = [
        PACMAN, '-Sy', 'archlinux-keyring', '--needed', '--noconfirm',
        '--disable-download-timeout'
    ]
    command = sudo(*command)
    command = ssh(system, *command, no_stdin=args.no_stdin)
    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def upgrade_system(system: int, args: Namespace) -> CompletedProcess:
    """Upgrades the system."""

    command = [PACMAN, '-Syu', '--needed', '--disable-download-timeout']

    for package in args.install:
        command.append(package)

    for glob in args.overwrite:
        command.append('--overwrite')
        command.append(glob)

    if args.yes:
        command = 'yes | ' + ' '.join(sudo(*command))
        command = ssh(system, command, no_stdin=args.no_stdin)
    else:
        command.append('--noconfirm')
        command = sudo(*command)
        command = ssh(system, *command, no_stdin=args.no_stdin)

    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def cleanup_system(system: int, args: Namespace) -> CompletedProcess:
    """Cleans up the system."""

    command = [PACMAN, '-Rncs', '$(pacman -Qmq)', '$(pacman -Qdtq)']

    if args.yes:
        command = 'yes | ' + ' '.join(sudo(*command))
        command = ssh(system, command, no_stdin=args.no_stdin)
    else:
        command.append('--noconfirm')
        command = sudo(*command)
        command = ssh(system, *command, no_stdin=args.no_stdin)

    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def upgrade(system: int, args: Namespace):
    """Upgrade process function."""

    syslogger(system).info('Upgrading system.')
    result = {}

    if args.keyring:
        completed_process = upgrade_keyring(system, args=args)
        result['keyring'] = completed_process_to_json(completed_process)

        if completed_process.returncode != 0:
            lograise(system, 'Could not update keyring.', completed_process)

    completed_process = upgrade_system(system, args=args)
    result['sysupgrade'] = completed_process_to_json(completed_process)

    if completed_process.returncode != 0:
        lograise(system, 'Could not upgrade system.', completed_process)

    if args.cleanup:
        completed_process = cleanup_system(system, args=args)
        result['pkgcleanup'] = completed_process_to_json(completed_process)

        if completed_process.returncode not in {0, 1}:
            lograise(system, 'Could not clean up system.', completed_process)

    return result


def sysupgrade(system: int, args: Namespace) -> bool:
    """Upgrated the respective system."""

    try:
        return upgrade(system, args)
    except SystemIOError as error:
        syslogger(system).error('I/O error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)
    except PacmanError as error:
        syslogger(system).error('Pacman error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)
    except UnknownError as error:
        syslogger(system).error('Unknown error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)



"""Common functions."""

from argparse import Namespace
from logging import DEBUG, INFO, WARNING
from subprocess import DEVNULL, PIPE, run, CompletedProcess
from typing import Iterable, List, Tuple, Union

from homeinfotools.rpc.common import HOSTNAME, SSH, SSH_OPTIONS, SUDO


__all__ = [
    'completed_process_to_json',
    'execute',
    'ssh',
    'sudo',
    'get_log_level'
]


def completed_process_to_json(completed_process: CompletedProcess) -> dict:
    """Converts a completed process into a JSON-ish dict."""

    return {
        'args': completed_process.args,
        'returncode': completed_process.returncode,
        'stdout': completed_process.stdout,
        'stderr': completed_process.stderr,
    }


def execute(command: Union[str, Iterable[str]]) -> CompletedProcess:
    """Executes the given command."""

    return run(command, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, text=True,
               check=False)


def ssh(system: int, *command: str, no_stdin: bool = False) -> List[str]:
    """Modifies the specified command to
    run via SSH on the specified system.
    """

    cmd = [SSH]

    if no_stdin:
        cmd.append('-n')

    for option in SSH_OPTIONS:
        cmd.append('-o')
        cmd.append(option)

    hostname = HOSTNAME.format(system)
    cmd.append(hostname)
    cmd.append(' '.join(command))
    return cmd


def sudo(*command: str) -> Tuple[str]:
    """Runs the command as sudo."""

    return (SUDO, ' '.join(command))


def get_log_level(args: Namespace) -> int:
    """Returns the set logging level."""

    return DEBUG if args.debug else INFO if args.verbose else WARNING



#! /usr/bin/env python3
"""Installation script."""

from setuptools import setup


setup(
    name='homeinfotools',
    version_format='{tag}',
    author='HOMEINFO - Digitale Informationssysteme GmbH',
    author_email='info@homeinfo.de',
    maintainer='Richard Neumann',
    maintainer_email='r.neumann@homeinfo.de',
    python_requires='>=3.8',
    install_requires=['setuptools-git-version', 'requests'],
    packages=[
        'homeinfotools',
        'homeinfotools.his',
        'homeinfotools.query',
        'homeinfotools.rpc',
        'homeinfotools.vpn'
    ],
    entry_points={
        'console_scripts': [
            'sysquery = homeinfotools.query.main:main',
            'sysrpc = homeinfotools.rpc.main:main',
            'sysvpn = homeinfotools.vpn.main:main',
        ],
    },
    license='GPLv3',
    description='Tools to manage HOMEINFO digital signge systems.'
)
